# This file is automatically generated by Hatch
from typing import Final

__all__ = ["__version__", "__version_tuple__"]

__version__: Final = '5.1.2'
__version_tuple__: tuple[int, int, int] = (5, 1, 2)
